import logging
import sys

from ...types import Timestamp
from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.deletedrevs")


class DeletedRevisions(Collection):
    PROPERTIES = {'ids', 'flags', 'timestamp', 'user', 'userid', 'size',
                  'sha1', 'contentmodel', 'comment', 'parsedcomment', 'content',
                  'tags'}

    # TODO:
    # This is *not* the right way to do this, but it should work for all queries.
    MAX_REVISIONS = 500

    def get(self, rev_id, *args, **kwargs):

        rev_id = int(rev_id)

        revs = list(self.query(revids={rev_id}, **kwargs))

        if len(revs) < 1:
            raise KeyError(rev_id)
        else:
            return revs[0]

    def query(self, *args, limit=sys.maxsize, **kwargs):
        """
        Queries deleted revisions.
        See https://www.mediawiki.org/wiki/API:Deletedrevs

        :Parameters:
            titles : set(str)
                A set of page names to query (note that namespace prefix is expected)
            start : :class:`mw.Timestamp`
                A timestamp to start querying from
            end : :class:`mw.Timestamp`
                A timestamp to end querying
            from_title : str
                A title from which to start querying (alphabetically)
            to_title : str
                A title from which to stop querying (alphabetically)
            prefix : str
                A title prefix to match on
            drcontinue : str
                When more results are available, use this to continue (3) Note: may only work if drdir is set to newer.
            unique : bool
                List only one revision for each page
            tag : str
                Only list revision tagged with this tag
            user : str
                Only list revisions saved by this user_text
            excludeuser : str
                Do not list revision saved by this user_text
            namespace : int
                Only list pages in this namespace (id)
            limit : int
                Limit the number of results
            direction : str
                "newer" or "older"
            properties : set(str)
                A list of properties to include in the results:


                * ids            - The ID of the revision.
                * flags          - Revision flags (minor).
                * timestamp      - The timestamp of the revision.
                * user           - User that made the revision.
                * userid         - User ID of the revision creator.
                * size           - Length (bytes) of the revision.
                * sha1           - SHA-1 (base 16) of the revision.
                * contentmodel   - Content model ID of the revision.
                * comment        - Comment by the user for the revision.
                * parsedcomment  - Parsed comment by the user for the revision.
                * content        - Text of the revision.
                * tags           - Tags for the revision.
        """
        # `limit` means something diffent here
        kwargs['limit'] = min(limit, self.MAX_REVISIONS)
        revisions_yielded = 0
        done = False
        while not done and revisions_yielded <= limit:
            rev_docs, query_continue = self._query(*args, **kwargs)
            for doc in rev_docs:
                yield doc
                revisions_yielded += 1
                if revisions_yielded >= limit:
                    break

            if query_continue != "" and len(rev_docs) > 0:
                kwargs['query_continue'] = query_continue
            else:
                done = True

    def _query(self, titles=None, pageids=None, revids=None,
               start=None, end=None, query_continue=None, unique=None, tag=None,
               user=None, excludeuser=None, namespace=None, limit=None,
               properties=None, direction=None):

        params = {
            'action': "query",
            'prop': "deletedrevisions"
        }

        params['titles'] = self._items(titles)
        params['pageids'] = self._items(pageids)
        params['revids'] = self._items(revids)
        params['drvprop'] = self._items(properties, levels=self.PROPERTIES)
        params['drvlimit'] = none_or(limit, int)
        params['drvstart'] = self._check_timestamp(start)
        params['drvend'] = self._check_timestamp(end)

        params['drvdir'] = self._check_direction(direction)
        params['drvuser'] = none_or(user, str)
        params['drvexcludeuser'] = none_or(excludeuser, int)
        params['drvtag'] = none_or(tag, str)
        params.update(query_continue or {'continue': ""})

        doc = self.session.get(params)
        doc_copy = dict(doc)

        try:
            if 'continue' in doc:
                query_continue = doc['continue']
            else:
                query_continue = ''

            pages = doc['query']['pages'].values()
            rev_docs = []

            for page_doc in pages:
                page_rev_docs = page_doc.get('deletedrevisions', [])

                try: del page_doc['deletedrevisions']
                except KeyError: pass

                for rev_doc in page_rev_docs:
                    rev_doc['page'] = page_doc

                rev_docs.extend(page_rev_docs)

            return rev_docs, query_continue

        except KeyError as e:
            raise MalformedResponse(str(e), doc)



from nose.tools import eq_

from ..namespace import Namespace


def test_namespace():
    namespace = Namespace(10, "Foo", canonical="Bar", aliases={'WT'},
                          case="foobar", content=False)

    eq_(namespace.id, 10)
    eq_(namespace.name, "Foo")
    eq_(namespace.canonical, "Bar")
    eq_(namespace.aliases, {'WT'})
    eq_(namespace.case, "foobar")
    eq_(namespace.content, False)

def test_namespace_from_doc():
    
    doc = {
        "id": 0,
        "case": "first-letter",
        "*": "",
        "content": ""
    }
    
    namespace = Namespace.from_doc(doc)
    eq_(namespace.id, 0)
    eq_(namespace.name, "")
    eq_(namespace.canonical, None)
    eq_(namespace.aliases, set())
    eq_(namespace.case, "first-letter")
    eq_(namespace.content, True)



RADIUS = 15
"""
TODO: Better documentation here.  For the time being, see:

Priedhorsky, R., Chen, J., Lam, S. T. K., Panciera, K., Terveen, L., &
Riedl, J. (2007, November). Creating, destroying, and restoring value in
Wikipedia. In Proceedings of the 2007 international ACM conference on
Supporting group work (pp. 259-268). ACM.
"""


class DUMMY_SHA1: pass
"""
Used in when checking for reverts when the checksum of the revision of interest
is unknown.

>>> DUMMY_SHA1 in {"aaa", "bbb"} # or any 40 character hex
False
>>>
>>> DUMMY_SHA1 == DUMMY_SHA1
True
>>> {DUMMY_SHA1, DUMMY_SHA1}
{<class '__main__.DUMMY_SHA1'>}
"""



import io
from multiprocessing import Queue

from nose.tools import eq_, raises

from ..processor import DONE, Processor


SAMPLE_XML = """
<mediawiki xmlns="http://www.mediawiki.org/xml/export-0.8/"
           xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
           xsi:schemaLocation="http://www.mediawiki.org/xml/export-0.8/
           http://www.mediawiki.org/xml/export-0.8.xsd"
           version="0.8" xml:lang="en">
  <siteinfo>
    <sitename>Wikipedia</sitename>
    <base>http://en.wikipedia.org/wiki/Main_Page</base>
    <generator>MediaWiki 1.22wmf2</generator>
    <case>first-letter</case>
    <namespaces>
      <namespace key="0" case="first-letter" />
      <namespace key="1" case="first-letter">Talk</namespace>
    </namespaces>
  </siteinfo>
  <page>
    <title>Foo</title>
    <ns>0</ns>
    <id>1</id>
    <revision>
      <id>1</id>
      <timestamp>2004-08-09T09:04:08Z</timestamp>
      <contributor>
        <username>Gen0cide</username>
        <id>92182</id>
      </contributor>
      <text xml:space="preserve">Revision 1 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
    <revision>
      <id>2</id>
      <timestamp>2004-08-10T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.109</ip>
      </contributor>
      <text xml:space="preserve">Revision 2 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <comment>Comment 2</comment>
      <format>text/x-wiki</format>
    </revision>
  </page>
  <page>
    <title>Bar</title>
    <ns>1</ns>
    <id>2</id>
    <revision>
      <id>3</id>
      <timestamp>2004-08-11T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.22</ip>
      </contributor>
      <text xml:space="preserve">Revision 3 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
  </page>
</mediawiki>"""



def test_processor():

    pathq = Queue()
    pathq.put(io.StringIO(SAMPLE_XML))

    outputq = Queue()

    def process_dump(dump, path):
        for page in dump:
            yield page.id


    processor = Processor(pathq, outputq, process_dump)
    processor.start()

    error, item = outputq.get()
    assert not error
    eq_(item, 1)

    error, item = outputq.get()
    assert not error
    eq_(item, 2)

    error, item = outputq.get()
    assert not error
    eq_(item, DONE)

def test_processor_error():

    pathq = Queue()
    pathq.put(io.StringIO(SAMPLE_XML))

    outputq = Queue()

    def process_dump(dump, path):
        raise Exception("foo")


    processor = Processor(pathq, outputq, process_dump)
    processor.start()

    error, item = outputq.get()
    assert error

    error, item = outputq.get()
    assert not error
    eq_(item, DONE)



class DummyChecksum():
    """
    Used in when checking for reverts when the checksum of the revision of interest
    is unknown.  DummyChecksums won't match eachother or anything else, but they
    will match themselves and they are hashable.

    >>> dummy1 = DummyChecksum()
    >>> dummy1
    <#140687347334280>
    >>> dummy1 == dummy1
    True
    >>>
    >>> dummy2 = DummyChecksum()
    >>> dummy2
    <#140687347334504>
    >>> dummy1 == dummy2
    False
    >>>
    >>> {"foo", "bar", dummy1, dummy1, dummy2}
    {<#140687347334280>, 'foo', <#140687347334504>, 'bar'}
    """
    
    def __str__(self): repr(self)
    def __repr__(self): return "<#" + str(id(self)) + ">"
